//
// ***** lcd7_main.c ***************
//
// Function: drive 4 x 7-seg LCD's of type NT-M5009B from Nelytech
//
// Processor:  18F8490
// Oscillator: 9.8304 MHz crystal x 4 (PLL)

// Heading of latest compiler listing:
// CCS PCH C Compiler, Version 4.018, 34261               09-May-10 10:51
//
//               Filename: lcd7_main.lst
//
//               ROM used: 5094 bytes (31%)
//                         Largest free fragment is 11286
//               RAM used: 239 (31%) at main() level
//                         333 (43%) worst case
//               Stack:    5 worst case (4 in main + 1 for interrupts)
//
// version-History:
// 100411- 1st implementation
// 100414- more robust (every LCD is updated every 10ms) and diagnostible (can read <dstr> on rs232)

// Communication is via I2C,
//  it's a slave with a 1-byte I2C device address 0xC2 or 0xC4, see file <lcd7.h>
//  and a 1-byte LCD address

// The I2C master first checks whether the device is ready
//  by reading 1 byte from device address and LCD address 0x80:
//  eg, send device address 0xc2, send LCD address 0x80, receive one byte
// If the received databyte is 0, the device is ready

// The I2C master sends the device address, LCD address and 5 data bytes for each LCD
// The LCD addresses are 0 for LCD#1, 5 for LCD#2, 10 for LCD#3, 15 for LCD#4
// eg, send device address 0xc2, send LCD address 0x00, send 20 data bytes for 4 LCD's

// the functions in the file <lcd7.c> convert the ASCII to LCD segments and drive the LCD's.

// All of the 4 LCD's can be updated in one message: device address, LCD addr=0, 4 x 5 = 20 data bytes
// Each group of 5 data bytes consists of 4 ASCII bytes = one byte for each digit, left to right,
//  and 1 byte with each bit || 7 | 6 | 5 | 4 || 3 | 2 | 1 | 0 || standing for a segment:
//  || "Battery" | "Continuity" | "-" | leftmost "1" || 5DP | 4DP | 3DP | 2DP ||
//  referring to the data sheet of the NT-M5009B from Nelytech

// the version with code_d2 = 1 also implements the UART for RS232
//  which allows some low level operations and diagnostics,
//  for the protocol see function <rx_msg> at the end of file <lcd7_utils.c>

#case

/************ version options **************/

#define code_d2         1

#if code_d2 == 0
 // I2C address = 0xC2
 #define c_version    0x80   // || RS232 | I2C only | - | - || - | - | - | - ||
#else
 // I2C address = 0xC4
 #define c_version    0x40
#endif

/*******************************************/

#nolist
 #include "18f8490.h"
// #fuses INTRC_IO,NOPROTECT,BROWNOUT,FCMEN,WDT16,BORV28,STVREN,NOMCLR // using int.osc.
 #fuses H4,NOPROTECT,BROWNOUT,FCMEN,WDT16,BORV28,STVREN,NOMCLR // using crystal osc.
//////// Fuses: LP,XT,HS,RC,EC,EC_IO,H4,RC_IO,PROTECT,NOPROTECT
//////// Fuses: BROWNOUT_NOSL,BROWNOUT_SW,NOBROWNOUT,BROWNOUT,WDT1,WDT2,WDT4
//////// Fuses: WDT8,WDT16,WDT32,WDT64,WDT128,WDT,NOWDT,BORV21,BORV28,BORV43
//////// Fuses: BORV46,PUT,NOPUT,IESO,NOIESO,NOSTVREN,STVREN,NODEBUG,DEBUG
//////// Fuses: FCMEN,NOFCMEN,INTRC_IO,INTRC,WDT256,WDT512,WDT1024,WDT2048
//////// Fuses: WDT4096,WDT8192,WDT16384,WDT32768,MCLR,NOMCLR,LPT1OSC
//////// Fuses: NOLPT1OSC,CCP2C1,CCP2E7,XINST,NOXINST,EBTR,NOEBTR
// #rom 0x300001 = {0x48, 0x16, 0x09}     // set the fuses
// #rom 0x300005 = {0x01, 0x81 }

 #include "reg8490.h"

#use fast_io(A)
#use fast_io(B)
//#use fast_io(C)
#use fast_io(D)
#use fast_io(E)
#use fast_io(F)
#use fast_io(G)
#use fast_io(H)
#use fast_io(J)

//#use delay(clock=39321600, restart_wdt) // use <32000000> with int.osc.&PLL
#use delay(clock=39321600, restart_wdt)

#if code_d2 == 1
#use rs232(baud=57600, xmit=PIN_C6, rcv=PIN_C7)
#endif

#include "lcd7.h"
//#use i2c(SLAVE, SDA=PIN_C4, SCL=PIN_C3, address=c_cop, force_hw, restart_wdt)

#list

/******************************************/

typedef union float4
{
  float f4;
  char f1[4];
};

typedef union Dat2
{ 
  int16 i;
  char c[2];
};

typedef union addrs
{
  int32 L;
  int16 i[2];
  char  c[4];
};

int16 fsr1;
#locate fsr1 = FSR1L            // covers 2 bytes: FSR1L and FSR1H

/********** globals **********/

char portB, pportB, portC, pportC, portE;       // PORT backup

#BIT bp_COM0      = portE.3     // out
#BIT bp_COM1      = portE.4     // out
#BIT bp_COM2      = portE.5     // out

/*** variables initialised to 0 ***/

char flags0;                    // || - | - | - || batt-lo | I2C err | FERR | OERR ||
#BIT b_oerr       = flags0.0
#BIT b_ferr       = flags0.1
#BIT b_eerr0      = flags0.2   // I2C R/W error
#BIT b_12V        = flags0.3   // battery-low condition

char sf1, sf2, sf3, sf4;
#BIT b_wdt        = sf1.1
#BIT b_1msec      = sf1.2       // 1ms toggle bit
#BIT b_10ms       = sf1.3
#BIT b_1sec       = sf1.4       // request to execute 1/sec functions
#BIT b_1min       = sf1.5       // request to refresh DTG from RTC
#BIT b_1day       = sf1.6

#BIT b_i2c        = sf2.0
#BIT b_cop        = sf2.1       // time to poll co-processor on I2C
#BIT b_pwrup      = sf2.2       // start after-pwr-up procedures
#BIT b_geti2c     = sf2.3       // ext.I2C device poll request
#BIT b_eerr       = sf2.4       // I2C R/W result not ok
#BIT b_232        = sf2.5       // identify msg.as coming from RS232
#BIT b_totl       = sf2.6
#BIT b_rxmsg      = sf2.7       // msg.timeout: request to process it

#BIT b_skeypresd  = sf3.0
#BIT b_lkeypresd  = sf3.1
#BIT b_skeyprssd  = sf3.2
#BIT b_lkeyprssd  = sf3.3
#BIT b_putprocs   = sf3.5
#BIT b_putcalcs   = sf3.6
#BIT b_putlimit   = sf3.7

#BIT b_AD         = sf4.0
#BIT b_AD1        = sf4.1
#BIT b_xfer       = sf4.2
#BIT b_xfermsg    = sf4.3
#BIT b_saveprm    = sf4.4

char teew;

// LCD display
char tdisp;

char dstr[20];
char updlcd;


#if code_d2 == 1
// RS232
char n_rxd;
int16 msglen;
char tx_in, tx_out;             /* RS232 tx buffer */
char rx_in, rx_out;             /* RS232 rx buffer */
char tm2m1;                     /* RS232 rx'd msg. timeout (10ms steps) */
#endif

// I2C
typedef enum {NOTHING, CONTROL, IDATA, ODATA} I2C_STATE;
I2C_STATE fState;
char i2caddr, ti2c;
char nxtrxi2c;

char rxi2c[c_rxi2c];
//  0: no.of char.s to be tx'd to I2C master
char txi2c[c_txi2c];
//  0: no.of char.s rx'd from I2C master

/********* other variables *********/

char rcon0;                     // || IPEN | - | - | nRI || nWDTO | nPD | nPOR | nBOR ||

char irq1;                      // exclusively for interrupt routines
#if code_d2 == 1
char irq2;
char irq_rs;
#endif

char tc10, t100;
char tpwrup, tcsec, t10sec;

#if code_d2 == 1
// RS232
char rxdata[c_rxdata];          // rx'd message on COM1
char tx[n_tx];                  /* RS232 tx buffer */
char rx[n_rx];                  /* RS232 rx buffer */
#endif

//char str[c_str];                // gen. string buffer

#list

/*******************************************************************/
/**************************   Utilities   **************************/
/*******************************************************************/

#INT_TIMER0
void TimrInt(void)
{	/* timer interrupt handler */

  /***********************/
  /* 500 usec operations */
  /***********************/

	bit_clear(intcon,TMR0IF);
	tmr0 += c_t_intr;
    tmr0h = 0;

    if (ti2c)
    { if (!--ti2c)
        fState = NOTHING;
    }

    /* update time-out counters (0,5ms steps) */

    if (b_1msec)
    { /*******************/
      /* 1 ms operations */
      /*******************/
      b_1msec = 0;

      if (teew) teew--;

      if (tdisp) tdisp--;
    }
    else
    { b_1msec  = 1;
      b_wdt = 1;                /* prove that interrupt is being executed */
    }

    if (!--tc10)
    { /********************/
	  /* 10 ms operations */
	  /********************/
      tc10 = c_t10;

     #if code_d2 == 1
      if (tm2m1)                // msg. timeout for rx'd msg.
      { if (!--tm2m1)
          b_rxmsg = 1;
      }
     #endif
      if (tpwrup)
      { if (!--tpwrup)
          b_pwrup = 1;
      }
    }

	if (!--t100)
	{ /*********************/
	  /* 100 ms operations */
	  /*********************/
      t100 = c_t100;            // reload 100 ms timer

      if (!--tcsec)
	  { /********************/
        /* 1 sec operations */
        /********************/
        tcsec = c_tsec;

        b_1sec = 1;

        if (!--t10sec)
        { /*** every 10 sec.s ***/
          t10sec = c_t10sec;
        }
      }
    }
}
/*******************************************************************/

#if code_d2 == 1

#INT_RDA
void rs232_rx(void)
{
  // check UART rx buffer and put into local rx buffer

    irq_rs = 0;
    if (bit_test(rcsta1,OERR))
    { b_oerr = 1;
      irq_rs = 1;
    }

    if (bit_test(rcsta1,FERR))
    { b_ferr = 1;
      irq_rs |= 2;
    }

    irq2 = 2;
    while (irq2)
    { 
      if (bit_test(pir1,RC1IF))
      { irq1 = rcreg1;
//        rx[rx_in] = rcreg1;       /* read the Rx'd byte into rx buffer - uses ind.reg.<INDF0> */
        fsr1 = rx + rx_in;          // = &rx[0] + rx_in = addr to be written, forcing the use of specified indirect registers
        indf1 = irq1;               // write to rx [rx_in ]

        if (!irq_rs)
        { irq1 = rx_in;
          irq1++;
          if (irq1 >= n_rx)
            irq1 = 0;
          if (irq1 != rx_out)       /* loose the latest byte if buffer is full */
            rx_in = irq1;
        }
      }
      irq2--;
    }

    if (irq_rs)
    { // restore receiver from error
      bit_clear(rcsta1,CREN);
      delay_us(2);
      bit_set(rcsta1,CREN);
      irq_rs = rcreg1;              // dummy read
    }
}
#endif
/*******************************************************************/

#INT_SSP
void ssp_interupt ()
{ // I2C handling

#BIT b_RnW    = sspstat.RnW
#BIT b_bf     = sspstat.BF

  char c1, sspstatm, lfsr0l, lfsr0h;

    sspstatm = sspstat & 0x2d;

    if ((sspstatm & 0x0c) == 0x0c)
    { // states 3 & 4: read operation (last byte was an addr or data)

      if (bit_test(sspstatm,DnA))
        while (b_bf);               // data: wait till previous byte is out
      else
        c1 = sspbuf;                // addr: dummy read (= rx'd addr)
      bit_clear(sspcon1,WCOL);

      sspbuf = nxtrxi2c;            // had to get here as quick as possible
      while (bit_test(sspcon1,WCOL))
      { bit_clear(sspcon1,WCOL);
        sspbuf = nxtrxi2c;
      }
      bit_set(sspcon1,CKP);         // terminate clock stretch (always required)

      lfsr0l = fsr0l;
      lfsr0h = fsr0h;

      if (i2caddr && (i2caddr < c_rxi2c-1))
      { i2caddr++;
        nxtrxi2c = rxi2c[i2caddr];
      }

      fState = ODATA;
      ti2c = c_ti2c;
    }
    else
    {
      lfsr0l = fsr0l;
      lfsr0h = fsr0h;

      if (b_bf)
        c1 = sspbuf;
      else
        c1 = 0;

      bit_clear(sspcon1,SSPOV);
      bit_clear(sspcon1,WCOL);

      if (sspstatm == 0x09)
      { // state 1: write operation, last byte was an addr, buffer is full
        fState = CONTROL;
        ti2c = c_ti2c;
      }
      else if (sspstatm == 0x29)
      { // state 2: write operation, last byte was data
        if (fState == CONTROL)
        { fState = IDATA;
          i2caddr = c1;

          if (i2caddr < c_rxi2c)
            nxtrxi2c = rxi2c[i2caddr];
          else
            nxtrxi2c = txi2c[0];
        }
        else
        { ti2c = c_ti2c;
          if (i2caddr < c_txi2c)
          { txi2c[i2caddr] = c1;
            if (i2caddr == 0)         // accept only 1 byte for addr.0 (protocol specific)
            { fState = NOTHING;
              ti2c = 0;
            }
            else
            { i2caddr++;
              if (i2caddr >= c_txi2c)
                i2caddr = c_txi2c - 1;
          } }
          else
          { rxi2c[0] = 0;            // acknowledge all rx bytes rx'd
            fState = NOTHING;
        } }
      }
    }
    bit_clear(pir1,SSPIF);
    fsr0l = lfsr0l;
    fsr0h = lfsr0h;
}
/**********************************************************************/

#INT_EXT1
void IB1(void)
{
  irq1 = PortB;             // dummy read
  intcon3 = 0xc0;
}
/**********************************************************************/

#include <lcd7.c>
#include <lcd7_utils.c>

/**********************************************************************/

void init(void)
{ /* very first start-up initialisation */

  char var1;
  char *pntr1;

    osccon = 0x70;          // also select primary oscillator when using int.osc.with PLL
    rcon0 = rcon;
    rcon = 0x53;
    bit_set(osctune,PLLEN); // for using int.osc.with PLL

    restart_wdt();

    PortA = 3;
    PortB = 1;
    PortC = 0xdb;
    PortD = 0;
    PortE = 0x40;
    PortF = 0;
    PortG = 0;
    PortH = 0;
    PortJ = 0;


    set_tris_a(0xc0);
    PortA = 3;
    set_tris_b(0xc0);
    PortB = 1;
    set_tris_c(0x98);
    PortC = 0xdb;
    set_tris_d(0);
    PortD = 0;
    set_tris_e(0);
    PortE = 0x40;
    set_tris_f(0);
    PortF = 0;
    set_tris_g(0x20);
    PortG = 0;
    set_tris_h(0);
    PortF = 0;
    set_tris_j(0);
    PortJ = 0;

    portB  = 1;
    pportB = 1;
    portC  = 0xdb;
    pportC = 0xdb;
    portE  = 0x40;

    t0con   = 0xc4;         // TMR0:8-bit, RTCC_INTERNAL, RTCC_DIV_32 -> 3.2552us/tick (4 us/tick with int.osc&PLL)

    var1 = 1;   //***
    // initialise variables
//    memset(&flags0, 0, &rcon0 - &flags0);
    pntr1 = &flags0;
    while (pntr1 < &rcon0)
    { *pntr1 = 0;
      pntr1++;
    }

    tc10        = c_t10 + 3;    // preset 10ms-timer
    t100        = c_t100 + 2;   // preset 100ms-timer
    tcsec       = c_tsec + 1;   // preset 1 sec timer
    t10sec      = c_t10sec;
    tpwrup      = c_tpwrup;
    tdisp       = 255;

    // LCD setup
    //lcdps = 0x00;               // || A/B | (1/3)/(1/2)bias | active | WRen || prescaler ||
    lcdse0 = 0xff;
    lcdse1 = 0xff;
    lcdse2 = 0xff;
    lcdse3 = 0xff;
    lcdse4 = 0xff;
    lcdse5 = 0xff;
    lcdcon = 0x02;
    bit_clear(lcdcon,WERR);
//    while (bit_test(lcdps,WA)) restart_wdt();   // synchronise to LCD frame
    bit_clear(pir3,LCDIF);
    bit_set(lcdcon,LCDEN);
    while (!bit_test(lcdps,WA)) restart_wdt();
    memset(lcddat1, 0xff, 11);  // full blast to show some life
    memset(lcddat2, 0xff, 7);   // full blast to show some life

// #use i2c(SLAVE, SDA=PIN_C4, SCL=PIN_C3, address=c_cop, force_hw, restart_wdt)
    // setup I2C
    sspadd  = c_cop & 0xfe;    // SPP's I2C address
    sspcon1 = c_sspcon1;

    //;  setup timing
    tmr0 = c_t_intr;

    //; enable interrupts
    bit_clear(pir1,ADIF);       // reset the result flag
    intcon = 0xe0;              // global IE: hi & lo, TMR0-IE, no INTE-B0intrpt
    bit_set(pie1,RC1IE);        // enable USART1 interrupt
    enable_interrupts(INT_SSP); // enable I2C

    var1 = 20;                  // time for >= 2 interrupts
    while (var1)
    { delay_us(84);
      restart_wdt();
      portC = PortC;
      var1--;
    }

    b_wdt = 1;
    watchdog();
}
/*************************************************************************/
/*************************************************************************/

void main(void)
{
  char mv1;
  union Dat2 mi1, mi2;

    init();

    // startup message
    strcpy(dstr+16,"chnl");
    dstr[15] = 0;
   #if code_d2 == 0
    strcpy(dstr+11,"1-4 ");
   #else
    strcpy(dstr+11,"5-8 ");
   #endif
    dstr[10] = 0;
    bin2bcd(dstr+6, c_version);
    bin2bcd(dstr+8, c_date1);
    dstr[5] = 0;
    bit_set(dstr[5], 1);        // DP3
    bin2bcd(dstr+1, c_date2);
    bin2bcd(dstr+3, c_date3);
    dstr[0] = 0;
    bit_set(dstr[0], 1);        // DP3

    updlcd = 0x0f;
    tdisp = c_tdisp1;

   #if code_d2 == 1
    tx[0] = '#';
    tx[1] = c_version;
    tx[2] = c_date1;
    tx[3] = c_date2;
    tx[4] = c_date3;

    tx_in = 5;
    tx_1();
   #endif

    /*** MAIN LOOP ***/
    for (;;)
    {
      if (b_pwrup)
      { // once-off pwrup procedures (late init.s)
        b_pwrup = 0;
      }

      if (!tpwrup)
      {
        if (!tdisp && bit_test(lcdps,WA))
        { tdisp = c_tdisp;

          updateDisplay();
          if ((updlcd & 0x0f) == 0)
            updlcd = 0x0f;
      } }

      /*** process I2C incoming data ***/
      if (txi2c[0])                     // = no.of data bytes following
      { // there's new data to process
        if (txi2c[0] > c_txi2c)
          txi2c[0] = 0;
        if (txi2c[0] > 1)
        { for (mi1.c[0]=1; mi1.c[0]<=txi2c[0]; mi1.c[0]++)
          { mv1 = mi1.c[0];
            mv1--;

            mi2.c[0] = txi2c[mi1.c[0]];
            dstr[mv1] = mi2.c[0];
        } }
        txi2c[0] = 0;                   // ack.receipt
      }

     #if code_d2 == 1
      /*** COM 1 ***/
      if (rx_in != rx_out)
      { // RS232 receive buffer contains something: process it
        b_gieh = 0;             /*** disable global interrupts ***/
        b_giel = 0;
        mi1.c[0] = rx[rx_out];
        rx_out++;
        if (rx_out >= n_rx)
          rx_out = 0;
        b_gieh = 1;             /*** enable global interrupts ***/
        b_giel = 1;

        if (msglen < c_rxdata)
        { rxdata[msglen] = mi1.c[0];
          msglen++;
          tm2m1 = c_tm2m;
        }
      }
      if (!tm2m1 && msglen)
      { b_232 = 1;
        rx_msg(rxdata, msglen);
        msglen = 0;
      }

      /*** COM1 output ***/
      tx_1();
     #endif

      watchdog();
    } // while (1)
}

/**********************************************************************/

